#-----------------------------------------------------------------------------
#
#   Supersymmetry - Particle detection
#
#-----------------------------------------------------------------------------

from math import pi, sin, cos
from random import choice, uniform, normalvariate, randrange, randint
from pygame import Color, Surface, draw, SRCALPHA
from albow.widget import Widget
from albow.utils import blit_in_rect

#-----------------------------------------------------------------------------

MAX_DEPTH = 3

MEAN_TRACE_LENGTH = 0.2
SD_TRACE_LENGTH = 0.1

MEAN_DV = 0.0
SD_DV = 0.4

CURVE_STEP_SIZE = 2.0

MEAN_CURVE_ACCEL = 0.0
SD_CURVE_ACCEL = 0.03

BG_COLOR = Color("black")
TRACE_COLOR = Color("cyan")

def int_point(x, y):
	return int(round(x)), int(round(y))

def random_unit_vector():
	a = uniform(0, 2 * pi)
	return cos(a), sin(a)

def split_trace(surf, x, y, vx, vy, depth):
	points = []
	n = randint(1, 2) + randint(1, 2)
	for i in xrange(n):
		dv = normalvariate(MEAN_DV, SD_DV)
		dvx = -vy * dv
		dvy = vx * dv
		points.append(gen_trace(surf, x, y, vx + dvx, vy + dvy, depth))
	#print "detection.split_trace: points =", points ###
	return choice(points)

def straight_trace(surf, x0, y0, vx, vy, length, depth):
	x1 = x0 + length * vx
	y1 = y0 + length * vy
	p0 = int_point(x0, y0)
	p1 = int_point(x1, y1)
	draw.line(surf, TRACE_COLOR, p0, p1)
	if surf.get_rect().collidepoint(p1) and depth < MAX_DEPTH:
		return split_trace(surf, x1, y1, vx, vy, depth)
	else:
		return p0

def curved_trace(surf, x0, y0, vx, vy, length, depth):
	p0 = int_point(x0, y0)
	p00 = p0
	s = CURVE_STEP_SIZE
	a = normalvariate(MEAN_CURVE_ACCEL, SD_CURVE_ACCEL)
	while length > 0.0:
		x1 = x0 + s * vx
		y1 = y0 + s * vy
		p1 = int_point(x1, y1)
		draw.line(surf, TRACE_COLOR, p0, p1)
		if not surf.get_rect().collidepoint(p1):
			return p00
		x0 = x1; y0 = y1; p0 = p1
		ax = -a * vy
		ay = a * vx
		vx += ax
		vy += ay
		length -= s
	if depth < MAX_DEPTH:
		return split_trace(surf, x0, y0, vx, vy, depth)
	else:
		return p0

def gen_trace(surf, x, y, vx, vy, depth):
	depth += 1
	#if depth <= MAX_DEPTH:
	length = surf.get_width() * normalvariate(MEAN_TRACE_LENGTH, SD_TRACE_LENGTH)
	if randint(0, 2) == 0:
		return straight_trace(surf, x, y, vx, vy, length, depth)
	else:
		return curved_trace(surf, x, y, vx, vy, length, depth)

def gen_image(surf):
	surf.fill(BG_COLOR)
	rect = surf.get_rect()
	w, h = surf.get_size()
	n = randint(1, 6) + randint(1, 6)
	points = []
	for i in xrange(n):
		vx, vy = random_unit_vector()
		points.append(gen_trace(surf, 0.5 * w, 0.5 * h, vx, vy, 0))
	#print "detection.gen_image: points =", points ###
	return choice(points)

#-----------------------------------------------------------------------------

class DetectorView(Widget):

	border_width = 2
	margin = 4
	bg_color = Color("white")
	circle_color = Color("yellow")
	circle_radius = 10

	def __init__(self, game, always_visible = False, **kwds):
		self.game = game
		self.always_visible = always_visible
		Widget.__init__(self, **kwds)
		m = 2 * self.margin
		if not game.detector_image:
			game.detector_image = Surface((self.width - m, self.height - m), SRCALPHA)
		#self.generate()
	
	def get_visible(self):
		return self.always_visible or self.game.world.running
	
	def draw(self, surf):
		blit_in_rect(surf, self.game.detector_image, surf.get_rect(), 'c')
		if self.always_visible:
			p = self.game.detector_point
			if p:
				m = self.margin
				draw.circle(surf, self.circle_color, (p[0] + m, p[1] + m), self.circle_radius, 1)

#	def generate(self):
#		gen_image(self.image)
#	
#	def mouse_down(self, e):
#		self.generate()
