#-----------------------------------------------------------------------------
#
#   Supersymmetry - Tiles
#
#-----------------------------------------------------------------------------

from pygame import Rect
from albow.resource import get_image
from albow.utils import blit_in_rect
from constants import *
from vectors import add, opposite_side

next_tileno = 1
tile_list = [None]

class Tile(object):
	#  tileno           int
	#  image            Image
	#  image_alignment  string
	#  cost             int
	#  rot_next         Tile
	#  routes           [(side, side)]
	
	fixed = False  # Cannot be moved by player
	routes = []
	helium_required = 0
	helium_provided = 0

	def __init__(self, image_spec, cost):
		global next_tileno
		self.tileno = next_tileno
		next_tileno += 1
		tile_list.append(self)
		image_file, self.image_alignment = image_spec
		self.image = get_image(image_file)
		self.cost = cost
		self.rot_next = self
	
	def exit_for_entrance(self, dir):
		rid = opposite_side(dir)
		for pair in self.routes:
			for i in (0, 1):
				if pair[i] == rid:
					return pair[1 - i]
		return SIDE_NONE
	
	def exits(self):
		for pair in self.routes:
			for dir in pair:
				if dir <> SIDE_NONE:
					yield dir
	
	def has_exit(self, dir):
		for exit in self.exits():
			if exit == dir:
				return True
		return False
		
	def draw(self, surf, x, y):
		s = TILE_SIZE
		r = Rect(x, y, s, s)
		self.draw_centered(surf, r)
	
	def draw_centered(self, surf, rect):
		blit_in_rect(surf, self.image, rect, self.image_alignment)

	def draw_connections(self, surf, x, y, cell, world):
		pass


class Pipe(Tile):

	def __init__(self, image_spec, routes, cost):
		Tile.__init__(self, image_spec, cost)
		self.routes = routes

#		if self.side1 == rid:
#			return self.side2
#		elif self.side2 == rid:
#			return self.side1
#		else:
#			return SIDE_NONE


class CornerPipe(Pipe):

	helium_required = 1


class Cap(Tile):
	#  exit  side

	def __init__(self, image_spec, exit, cost):
		Tile.__init__(self, image_spec, cost)
		self.exit = exit
	
	def exits(self):
		yield self.exit


class TeePipe(Pipe):

	def __init__(self, image_spec, through, side, cost):
		Tile.__init__(self, image_spec, cost)
		self.routes = [through, (side, SIDE_NONE)]


class FixedPipe(Pipe):

	fixed = True

class Obstacle(Tile):
	
	fixed = True
	
class Helium(Tile):
	
	fixed = True
	helium_provided = 20
	
	connection_images = [
		(get_image("helium-e.png"), 'r'),
		(get_image("helium-s.png"), 'b'),
		(get_image("helium-w.png"), 'l'),
		(get_image("helium-n.png"), 't'),
	]
	
	def has_exit(self, dir):
		return True
	
	def draw_connections(self, surf, x, y, cell, world):
		s = TILE_SIZE
		r = Rect(x, y, s, s)
		for dir in xrange(4):
			tile = world.get_tile(add(cell, direction_vectors[dir]))
			if tile and tile.has_exit(opposite_side(dir)):
				image, align = self.connection_images[dir]
				blit_in_rect(surf, image, r, align)

#-----------------------------------------------------------------------------

def define_rotset(seq):
	n = len(seq)
	for i, tile in enumerate(seq):
		tile.rot_next = seq[(i + 1) % n]
	return seq

def rotate_tile(tile):
	if tile:
		return tile.rot_next

def get(tileno):
	return tile_list[tileno]

def tileno(tile):
	if tile:
		return tile.tileno
	else:
		return 0

#-----------------------------------------------------------------------------

PIPE_EW = Pipe(("pipe-ew.png", 'c'), [(SIDE_E, SIDE_W)], PIPE_COST)
PIPE_NS = Pipe(("pipe-ns.png", 'c'), [(SIDE_N, SIDE_S)], PIPE_COST)
PIPE_SW = CornerPipe(("pipe-sw.png", 'bl'), [(SIDE_S, SIDE_W)], CORNER_COST)
PIPE_NW = CornerPipe(("pipe-nw.png", 'tl'), [(SIDE_N, SIDE_W)], CORNER_COST)
PIPE_NE = CornerPipe(("pipe-ne.png", 'tr'), [(SIDE_N, SIDE_E)], CORNER_COST)
PIPE_SE = CornerPipe(("pipe-se.png", 'br'), [(SIDE_S, SIDE_E)], CORNER_COST)
PIPE_CROSS = Pipe(("pipe-nsew.png", 'c'), [(SIDE_N, SIDE_S), (SIDE_E, SIDE_W)], CROSS_COST)
PIPE_START = FixedPipe(("pipe-start.png", 'r'), [(SIDE_N, SIDE_S), (SIDE_E, SIDE_NONE)], 0)
PIPE_N = Cap(("cap-n.png", 't'), SIDE_N, CAP_COST)
PIPE_S = Cap(("cap-s.png", 'b'), SIDE_S, CAP_COST)
PIPE_E = Cap(("cap-e.png", 'r'), SIDE_E, CAP_COST)
PIPE_W = Cap(("cap-w.png", 'l'), SIDE_W, CAP_COST)
OBSTACLE = Obstacle(("obstacle.png", 'c'), 0)
HELIUM = Helium(("helium_plant.png", 'c'), 0)
TEE_N = TeePipe(("tee-n.png", 't'), (SIDE_E, SIDE_W), SIDE_N, TEE_COST)
TEE_S = TeePipe(("tee-s.png", 'b'), (SIDE_E, SIDE_W), SIDE_S, TEE_COST)
TEE_E = TeePipe(("tee-e.png", 'r'), (SIDE_N, SIDE_S), SIDE_E, TEE_COST)
TEE_W = TeePipe(("tee-w.png", 'l'), (SIDE_N, SIDE_S), SIDE_W, TEE_COST)

straight_pipes = define_rotset([PIPE_EW, PIPE_NS])
cap_pipes = define_rotset([PIPE_N, PIPE_E, PIPE_S, PIPE_W])
corner_pipes = define_rotset([PIPE_SW, PIPE_NW, PIPE_NE, PIPE_SE])
tee_pipes = define_rotset([TEE_N, TEE_E, TEE_S, TEE_W])
random_pipes = 20 * [PIPE_EW] + 3 * [PIPE_SW] + [PIPE_CROSS]
buyable_pipes = [PIPE_EW, PIPE_SW, TEE_N, PIPE_CROSS, PIPE_W]

