#-----------------------------------------------------------------------------
#
#   Supersymmetry - World
#
#-----------------------------------------------------------------------------

from __future__ import division
import random, traceback
from math import ceil
from pygame import Rect
from albow.resource import get_sound
from albow.dialogs import alert, ask
from constants import *
from vectors import add, opposite_side
import tiles
import particles

beam_sound = get_sound("beam.ogg")
beam_sound.set_volume(0.4)

switch_on_sound = get_sound("switch-on.ogg")
switch_off_sound = get_sound("switch-off.ogg")
success_sound = get_sound("success.ogg")
skint_sound = get_sound("skint.ogg")
money_sound = get_sound("money.ogg")
game_over_sound = get_sound("applause.ogg")
champagne_sound = get_sound("champagne.ogg")
placement_sound = get_sound("place.ogg")
removal_sound = get_sound("remove.ogg")
rotation_sound = get_sound("rotate.ogg")

#-----------------------------------------------------------------------------

class World(object):
  #  level_number   int
  #  initial_grid   int[grid_size][grid_size]
	#  grid           int[grid_size][grid_size]
	#  start_cell     (col, row)
	#  initial_funds  int
	#  funds          int
	#  game_over      bool
	#  grid_bounds    Rect
	#  initial_inventory    [int]
	#  inventory            [int]
	#  initial_inventory_selection  int
	#  inventory_selection  int
	#  circuit_complete     bool
	#  vacuum_ok            bool
	#  beam_energy          int
	#  running              bool
	#  particles_found      int
	#  detection_time       float
	
	game_over = False
	initial_inventory = []
	vacuum_ok = False
	initial_extra_funds = EXTRA_FUNDS
	extra_funds = initial_extra_funds
	helium_level = 0

	def __init__(self):
		self.level_number = 0
		self.funds = STARTING_FUNDS
		self.grid = make_grid()
		self.game_over = False
		self.inventory = [] # [1, 3]
		self.inventory_selection = 0
		self.running = False
		self.particles_found = 0
		self.add_obstacles()
		self.start_cell = (0, GRID_ROWS // 2)
		self.place_start_pipe()
		self.add_helium()
		self.new_level()
		#self.dump_levels() ###
	
	def add_obstacles(self):
		tileno = tiles.OBSTACLE.tileno
		for i in xrange(random.randrange(MIN_OBSTACLES, MAX_OBSTACLES + 1)):
			x = random.randrange(GRID_COLS - 1)
			y = random.randrange(GRID_ROWS - 1)
			for dx in xrange(random.randrange(1, 3)):
				for dy in xrange(random.randrange(1, 3)):
					self.grid[y + dy][x + dx] = tileno
	
#	def add_helium(self):
#		tileno = tiles.HELIUM.tileno
#		for i in xrange(random.randrange(MIN_HELIUM, MAX_HELIUM + 1)):
#			x = random.randrange(2, GRID_COLS - 1)
#			y = random.randrange(GRID_ROWS - 1)
#			for dx in xrange(random.randrange(1, 3)):
#				for dy in xrange(random.randrange(1, 3)):
#					self.grid[y + dy][x + dx] = tileno
	
	def add_helium(self):
		tile = tiles.HELIUM
		for x0 in xrange(2, GRID_COLS, HELIUM_SPACING):
			for y0 in xrange(0, GRID_COLS, HELIUM_SPACING):
				x = x0 + random.randint(0, HELIUM_SPACING - 1)
				y = y0 + random.randint(0, HELIUM_SPACING - 1)
				self.set_tile((x, y), tile)

	def place_start_pipe(self):
		col, row = self.start_cell
		for x in xrange(col, col + 2):
			for y in xrange(row - 1, row + 2):
				self.grid[row][col] = 0
		self.grid[row][col] = tiles.PIPE_START.tileno
		self.grid[row][col + 1] = tiles.HELIUM.tileno
	
	def new_level(self):
		if self.particles_found >= particles.num_particles():
			self.game_over = True
			announce_success()
			return
		self.level_number += 1
		n = self.level_number
		grant = self.funding_grant_for_level(n)
		energy = self.energy_required_for_level(n)
		if self.level_number > 1:
			alert("The next phase of the project will require an energy of {} TeV. "
				"Additional funding of {:g} billion euros had been granted.".format(energy, grant * 1e-9))
		self.funds += grant
		self.initial_funds = self.funds
		self.initial_grid = self.grid
		self.initial_inventory = self.inventory
		self.initial_inventory_selection = self.inventory_selection
		self.extra_funds = int(self.extra_funds * INTEREST_RATE)
		self.report_extra_funds()
		self.initial_extra_funds = self.extra_funds
		self.grid_bounds = grid_bounds_for_level(self.level_number)
#		if self.level_number == 1:
#			self.start_cell = (0, GRID_ROWS // 2)
#			self.place_start_pipe()
		self.start_level()
	
	def generate_detection_time(self, scale):
		self.detection_time = random.normalvariate(
			scale * MEAN_DETECTION_TIME, scale * SD_DETECTION_TIME)
		
	def start_level(self):
		self.grid = copy_grid(self.initial_grid)
		self.funds = self.initial_funds
		self.inventory = self.initial_inventory[:]
		self.inventory_selection = self.initial_inventory_selection
		self.extra_funds = self.initial_extra_funds
		self.generate_detection_time(1.0)
	
	def restart_level(self):
		if ask("Restart the current level?") == "OK":
			self.start_level()
	
	def get_tile(self, cell):
		x, y = cell
		if 0 <= x < GRID_COLS and 0 <= y < GRID_ROWS:
			return tiles.get(self.grid[y][x])
	
	def set_tile(self, cell, tile):
		x, y = cell
		if 0 <= x < GRID_COLS and 0 <= y < GRID_ROWS:
			self.grid[y][x] = tiles.tileno(tile)
			game.changed()
	
	def cell_in_bounds(self, cell):
		return self.grid_bounds.collidepoint(cell)
	
	def place_tile(self, cell):
		tile = self.get_tile(cell)
		if tile:
			if tile.fixed:
				return
			tile = tiles.rotate_tile(tile)
			self.set_tile(cell, tile)
			rotation_sound.play()
		else:
			if not self.have_funds(PLACEMENT_COST):
				return
			#print "TODO: World.place_tile: Play placement sound" ###
			tile = self.take_tile_from_inventory()
			if not tile:
				return
			self.spend_funds(PLACEMENT_COST)
			self.set_tile(cell, tile)
			placement_sound.play()
	
	def remove_tile(self, cell):
		tile = self.get_tile(cell)
		if tile and not tile.fixed and self.room_in_inventory():
			if self.spend_funds(REMOVAL_COST):
				self.add_tile_to_inventory(tile)
				self.set_tile(cell, None)
				removal_sound.play()
	
	def room_in_inventory(self):
		return True
	
	def take_tile_from_inventory(self):
		inventory = self.inventory
		i = self.inventory_selection
		n = len(inventory)
		if 0 <= i < n:
			tileno = inventory.pop(i)
			n -= 1
			if i >= n:
				i = n - 1
			if n and inventory[i] <> tileno:
				if tileno in inventory:
					i = inventory.index(tileno)
			self.inventory_selection = i
			game.changed()
			return tiles.get(tileno)
	
	def add_tile_to_inventory(self, tile):
		if tile:
			inventory = self.inventory
			i = self.inventory_selection
			n = len(inventory)
			tileno = tile.tileno
			if tileno in inventory:
				j = inventory.index(tileno)
				while j < n and inventory[j] == tileno:
					j += 1
			else:
				j = len(inventory)
			inventory.insert(j, tileno)
			self.inventory_selection = j
			game.changed()

	def have_funds(self, amount):
		if amount <= self.funds:
			return True
		else:
			self.stop()
			skint_sound.play()
			alert("Out of funds!")
			return False
	
	def spend_funds(self, amount):
		if self.have_funds(amount):
			self.funds -= amount
			game.changed()
			return True
		else:
			return False
	
	def spend_funds_with_sound(self, amount):
		result = self.spend_funds(amount)
		if result:
			money_sound.play()
		return result
	
	def grant_extra_funds(self, amount):
		if amount < self.extra_funds:
			self.extra_funds -= amount
			self.funds += amount
			game.changed()
			money_sound.play()
			self.report_extra_funds()
	
	def report_extra_funds(self):
		if 0:
			print "Extra funds now {:,d}".format(self.extra_funds)
	
	def extra_funds_available(self):
		return self.extra_funds > 1000000000
	
	def buy_random_pipe(self):
		if self.room_in_inventory():
			tile = random.choice(tiles.random_pipes)
			if self.spend_funds(tile.cost):
				money_sound.play()
				self.add_tile_to_inventory(tile)

	def timestep(self, dt):
		self.update_status()
		if self.running:
			if self.able_to_run() and self.spend_funds(RUNNING_COST):
				self.run(dt)
			else:
				self.stop()
	
	def update_status(self):
		self.circuit_complete = self.trace_beam()
		self.helium_level = 0
		self.find_helium(self.start_cell, SIDE_N, set())
#		if not self.able_to_run():
#			self.stop()
	
	def trace_beam(self):
		start = (self.start_cell, SIDE_N)
		state = start
		self.beam_energy = 0
		self.vacuum_ok = True
		self.helium_required = 0
		while 1:
			self.beam_energy += PIPE_ENERGY
			if not self.check_vacuum(state[0]):
				self.vacuum_ok = False
			self.helium_required += self.get_tile(state[0]).helium_required
			state = self.step_beam(state)
			if not state:
				return False
			if state == start:
				return True
	
	def check_vacuum(self, cell):
		tile = self.get_tile(cell)
		for dir in tile.exits():
			neighbour_cell = add(cell, direction_vectors[dir])
			neighbour_tile = self.get_tile(neighbour_cell)
			if not neighbour_tile:
				return False
			rid = opposite_side(dir)
			if not neighbour_tile.has_exit(rid):
				return False
		return True
	
	def find_helium(self, cell, in_dir, visited):
		if cell not in visited:
			visited.add(cell)
			in_rid = opposite_side(in_dir)
			tile = self.get_tile(cell)
			if tile and tile.has_exit(in_rid):
				if tile.helium_provided:
					self.helium_level += tile.helium_provided
				else:
					for out_dir in tile.exits():
						neighbour_cell = add(cell, direction_vectors[out_dir])
						self.find_helium(neighbour_cell, out_dir, visited)
	
	def step_beam(self, state):
		cell, dir = state
		cell = add(cell, direction_vectors[dir])
		tile = self.get_tile(cell)
		if not tile:
			return None
		dir = tile.exit_for_entrance(dir)
		if dir == SIDE_NONE:
			return None
		return (cell, dir)
	
	def run(self, dt):
		if self.beam_energy >= self.energy_required:
			self.detection_time -= dt
			#particle = particles.get(self.particles_found)
			#if random.random() <= particle.discovery_chance:
			if self.detection_time <= 0.0:
				self.stop()
				success_sound.play()
				particle = particles.get(self.particles_found)
				particle.announce_discovery(game)
				self.particles_found += 1
				game.changed()
				if particle.is_real:
					self.new_level()
				else:
					self.generate_detection_time(0.5)
					self.start()
	
	@property
	def beam_status(self):
		if self.circuit_complete:
			return "Complete"
		else:
			return "Incomplete"
	
	@property
	def vacuum_status(self):
		if self.vacuum_ok:
			return "Yes"
		else:
			return "No"
	
	@property
	def helium_status(self):
		avail = self.helium_level
		reqd = self.helium_required
		return "%s / %s   (%+d)" % (avail, reqd, avail - reqd)
	
	@property
	def energy_required(self):
		return self.energy_required_for_level(self.level_number)
	
	def energy_required_for_level(self, n):
		if n == 0:
			return 0
		k = 0.4 + 0.1 * (n - 1) / 14
		w, h = grid_dimensions_for_level(n)
		return int(round(k * w * h))
	
	def funds_required_for_level(self, n):
		e = self.energy_required_for_level(n) - self.energy_required_for_level(n-1)
		p = ceil(e / PIPE_ENERGY)
		#s = grid_size_for_level(n)
		#c = 2 * ceil( p / s)
		a = (p - 4) * PIPE_COST + 3 * CORNER_COST + CROSS_COST + p * PLACEMENT_COST
		k = 1.05 # - 0.05 * (n - 1) / 14
		amount = a * k
		r = 500000000
		m = ceil(amount / r)
		return int(m) * r
	
	def funding_grant_for_level(self, n):
		required = self.funds_required_for_level(n)
		carried_forward = max(0, self.funds - MAX_CARRY_FORWARD)
		amount = max(0, required - carried_forward)
		if 0:
			print "Funds left from previous level: {:16,d}".format(self.funds)
			print "Funding calculated for level {:2d}: {:16,d}".format(n, required)
			print "Less carry forward:             {:16,d}".format(carried_forward)
			print "Net grant:                      {:16,d}".format(amount)
		return amount
	
	def dump_levels(self):
		mt = 0
		print "%5s %6s %15s" % ("Level", "Energy", "Funds")
		for n in xrange(1, 16):
			e = self.energy_required_for_level(n)
			m = int(self.funds_required_for_level(n))
			mt += m
			print "{:5d} {:6d} {:15,d}".format(n, e, m)
		print "{:5s} {:6s} {:15,d}".format("", "", mt)
	
	def able_to_run(self):
		#print "World.able_to_run", self.circuit_complete ###
		#traceback.print_stack()
		return self.circuit_complete and self.vacuum_ok and self.helium_ok
	
	@property
	def helium_ok(self):
		return self.helium_level >= self.helium_required
	
	def is_startable(self):
		return self.able_to_run() and not self.running
	
	def is_running(self):
		return self.running
	
	def start(self):
		if self.able_to_run():
			self.running = True
			game.changed()
			#print "TODO: World.start: Start playing running sound" ###
			switch_on_sound.play()
			beam_sound.play(loops = -1)
	
	def stop(self):
		if self.running:
			self.running = False
			game.changed()
			#print "TODO: World.start: Stop playing running sound" ###
			switch_off_sound.play()
			beam_sound.fadeout(100)

#-----------------------------------------------------------------------------

def make_grid():
	z = TILE_NONE
	return [[z] * GRID_COLS for i in xrange(GRID_ROWS)]

def copy_grid(grid):
	return [row[:] for row in grid]

def grid_bounds_for_level(n):
	w, h = grid_dimensions_for_level(n)
	return Rect(0, (GRID_ROWS - h) // 2, w, h)

def grid_dimensions_for_level(n):
	return level_zones[n - 1]

#	w = h = grid_size_for_level(n)
#	if h > GRID_ROWS:
#		#w += (GRID_ROWS - h) // 2
#		h = GRID_ROWS
#	if w > GRID_COLS:
#		w = GRID_COLS
#	return w, h

def grid_size_for_level(n):
	return LEVEL1_GRID_SIZE + 2 * n

def announce_success():
	game_over_sound.play()
	result = ask("Excellent! You've won a Nobel Prize for finding all the supersymmetric partners.",
		["Modestly Decline", "Open Champagne"])
	if result == "Open Champagne":
		champagne_sound.play()

#-----------------------------------------------------------------------------

level_zones = [
	(7, 7),
	(9, 9),
	(11, 11),
	(12, 13),
	(13, 15),
	(14, 17),
	(15, 19),
	(17, 19),
	(19, 19),
	(21, 19),
	(23, 19),
	(25, 19),
	(27, 19),
	(29, 19),
	(31, 19),
]
